# 11.代码模块化

- 使用模块模式
- 使用当前标准编写模块代码:AMD 和 CommonJS
- 使用 ES6 模块

## 在 JavaScript ES6 之前的版本中模块化代码

### 使用对象、闭包和立即执行函数实现模块

- 隐藏模块内部实现——我们已经知道，调用 JavaScript 函数创建新的作用域，我们可以在该作用域中定义变量，此时定义的变量只在当前函数中可见。因此，隐藏模块内部实现的一个选择是使用函数作为模块。采用这种方式，所有函数变量都成为模块内部变量，模块外部不可见。
- 定义模块接口——使用函数实现模块意味着只能在模块内部访问变量。但是，如果使用其他代码调用该模块，我们必须定义简洁的接口，可以通过接口暴露模块提供的功能。一种实现方式是利用对象和闭包。思路是，通过函数模块返回代表模块公共接口的对象。该对象必须包含模块提供的方法，而这些方法将通过闭包保持模块内部变量，甚至在模块函数执行完成之后仍然保持模块变量。

**使用函数作为模块**

```js
(function countClicks() {
  let numClicks = 0;
  document.addEventListener('click', () => {
    alert(++numClicks);
  });
})();
```

**模块模式:使用函数扩展模块，使用对象实现接口**

```js
const MouseCounterModule = (function() {
  let numClicks = 0;
  const handleClick = () => {
    alert(++numClicks);
  };
  return {
    countClicks: () => {
      document.addEventListener('click', handleClick);
    }
  };
})();
assert(
  typeof MouseCounterModule.countClicks === 'function',
  'We can access module functionality'
);
assert(
  typeof MouseCounterModule.numClicks === 'undefined' &&
    typeof MouseCounterModule.handleClick === 'undefined',
  'We cannot access internal module details'
);
```

**模块扩展**

```js
const MouseCounterModule = (function() {
  let numClicks = 0;
  const handleClick = () => {
    alert(++numClicks);
  };
  return {
    countClicks: () => {
      document.addEventListener('click', handleClick);
    }
  };
})();
(function(module) {
  // 立即调用一个函数，该函数接收需要扩展的模块作为参数
  let numScrolls = 0;
  const handleScroll = () => {
    alert(++numScrolls);
  };
  module.countScrolls = () => {
    // 扩展模块接口
    document.addEventListener('wheel', handleScroll);
  };
})(MouseCounterModule);
assert(
  typeof MouseCounterModule.countClicks === 'function',
  'We can access initial module functionality'
);
assert(
  typeof MouseCounterModule.countScrolls === 'function',
  'We can access augmented module functionality'
);
```

扩展模块的过程与重新创建模块的过程类似。调用立即执行函数，并传入需要扩展的模块作为参数:

```js
(function(module){
  ...
  return module;
})(MouseCounterModule);
```

在函数内，增加我们所需的变量和函数。在本例中，我们定义一个私有变量和一个私有函数，计算并显示滚动次数:

```js
let numScrolls = 0;
const handleScroll = () => {
  alert(++numScrolls);
};
```

最后，通过立即执行函数的 module 参数扩展模块，这与扩展其他对象的过程类似:

```js
module.countScrolls = () => {
  document.addEventListener('wheel', handleScroll);
};
```

执行这个简单操作之后，MouseCounterModule 也可以调用 countScrolls 方法。 现在模块公共接口有两个方法，我们可以这样使用模块:

```js
MouseCounterModule.countClicks();
MouseCounterModule.countScrolls();
```

> 通过独立的立即执行函数扩展模块，无法共享模块私有变量，因为每个函数都分别创建了新的作用域。虽然这是一个缺点，但并不致命，我们仍然可以使用模块模式保持 JavaScript 应用模块化。

注意到在模块模式中，模块就像对象一样，我们可以采用任何合适的方式进行扩展。例如，我们可以在模块上添加新属性:

```js
MouseCounterModule.newMethod = ()=> {...}
```

我们也可以使用相同的原则轻松地创建子模块:

```js
MouseCounterModule.newSubmodule = () => {
  return {...};
}();
```

所有这些方法都有模块模式相同的缺点:扩展的模块无法共享原有模块的内部属性。糟糕的是，模块模式还有其他问题。当我们开始创建模块化应用时，模块本身常常依赖其他模块的功能。然而，模块模式无法实现这些依赖关系。我们作为开发者，则不得不考虑正确的依赖顺序，这样我们的模块才具有执行时所需的完整的依赖。虽然在小型或中型应用中不是大问题，但是，在使用大量内部模块依赖的大型应用中则是非常严重的问题。

### 使用 AMD 和 CommonJS 模块化 JavaScript 应用

- AMD(RequireJS)

**使用 AMD 定义模块依赖于 jQuery**

```js
define('MouseCounterModule', ['jQuery'], $ => {
  let numClicks = 0;
  const handleClick = () => {
    alert(++numClicks);
  };
  return {
    countClicks: () => {
      $(document).on('click', handleClick);
    }
  };
});
```

AMD 有以下几项优点。

- 自动处理依赖，我们无需考虑模块引入的顺序。
- 异步加载模块，避免阻塞。
- 在同一个文件中可以定义多个模块。

- CommonJS

**使用 CommonJS 定义模块**

```js
//MouseCounterModule.js
const $ = require('jQuery');
let numClicks = 0;
const handleClick = () => {
  alert(++numClicks);
};
module.exports = {
  countClicks: () => {
    $(document).on('click', handleClick);
  }
};
```

在另一个文件中引用该模块，可以这样写:

```js
const MouseCounterModule = require("MouseCounterModule.js");
MouseCounterModule.countClicks();
```

CommonJS 具有两个优势。

- 语法简单。只需定义 `module.exports` 属性，剩下的模块代码与标准 JavaScript 无差异。引用模块的方法也很简单，只需要使用 `require` 函数。
- CommonJS 是 Node.js 默认的模块格式，所以我们可以使用 npm 上成千上万的包。

## ES6 模块

ES6 模块结合了 CommonJS 与 AMD 的优点，具体如下:

- 与 CommonJS 类似，ES6 模块语法相对简单，并且基于文件(每个文件就是一个模块)。
- 与 AMD 类似，ES6 模块支持异步模块加载。

> 内置模块是ES6标准的一部分。很快就可以看到，ES6模块语法包含语义、关键字(如 关键字 `export` 与 `import`)。目前部分浏览器尚未支持 ES6。如果现在就需要使用 ES6 模块，我们需要对代码进行编译，可以使用 [Traceur](https://github.com/google/traceur-compiler)，[Babel](http://babeljs.io/)或 [TypeScript](www.typescriptlang.org/)。我们还可以使用 [SystemJS library](https://github.com/systemjs/systemjs)，SystemJS支持目前所有的模块标准: AMD、CommonJS 甚至 ES6 模块。可以在 SystemJS [工程目录](https://github.com/systemjs/systemjs) 查看使用说明。

为了提供这个功能，ES6 引入两个关键字。

- export —— 从模块外部指定标识符。
- import —— 导入模块标识符。

**回顾ES6模块语法**

```js
export const ninja = "Yoshi"; // 导出变量
export function compare(){} // 导出函数
export class Ninja{} // 导出类
export default class Ninja{} // 导出默认类
export default function Ninja(){} // 导出默认函数
const ninja = "Yoshi";
function compare(){};
export {ninja, compare}; // 导出存在的变量
export {ninja as samurai, compare}; // 使用别名导出变量
import Ninja from "Ninja.js"; // 导入默认导出
import {ninja, Ninja} from "Ninja.js"; // 导入命名导出
import * as Ninja from "Ninja.js"; // 导入模块中声明的全部导出内容
import {ninja as iNinja} from "Ninja.js"; // 通过别名导入模块中声明的全部导出内容
```

## 小结

- 小的、组织良好的代码远比庞大的代码更容易理解和维护。优化程序结构和组织方式的一种方式是将代码拆分成小的、耦合相对松散的片段或模块。
- 模块是比对象或函数稍大的、用于组织代码的单元，通过模块可以将程序进行分类。
- 通常来说，模块可以降低理解成本，模块易于维护，并可以提高代码的可重用性。
- 在 JavaScript ES6 之前，没有内置的模块，开发者们不得不创造性地发挥 JavaScript 语言现有的特性实现模块化。最流行的方式之一是通过立即执行函数的闭包实现模块。
  - 使用立即执行函数创建定义模块变量的闭包，从外部作用域无法访问这些变量。
  - 使用闭包可以使模块变量保持活跃。
  - 最流行的是模块模式，通常采用立即执行函数，并返回一个新对象作为模块的公共接口。
- 除了模块模式，还有两个流行的模块标准:AMD，可以在浏览器端使用；CommonJS，在 JavaScript 服务端更流行。AMD 可以自动解决依赖，异步加载模块，避免阻塞。CommonJS 语法简单，可以同步加载模块(因此在服务端更流行)，通过 npm(node 包管理)可以获取大量模块。
- ES6 结合了 AMD 和 CommonJS 的特点。ES6 模块受 CommonJS 影响，语法简单，并提供了与 AMD 类似的异步模块加载机制。
  - ES6 模块基于文件，一个文件是一个模块。
  - 通过关键字 `export` 导出标识符，在其他模块中可引用这些标识符。
  - 在其他模块中通过关键字 `import` 导入标识符。
  - 模块可以使用默认导出，通过一个 `export` 导出整个模块。
  - `export` 与 `import` 都可以通过关键字 `as` 使用别名。