## 搭建基本的前端开发环境

- 构建发布需要的 HTML、CSS、JS 文件
- 使用 CSS 预处理器来编写样式
- 处理和压缩图片
- 使用 Babel 来支持 ES 新特性
- 本地提供静态服务以方便开发调试

### 关联 HTML

> 参考文档：[html-webpack-plugin](https://webpack.docschina.org/plugins/html-webpack-plugin/)

```sh
npm install html-webpack-plugin -D

# 或者
yarn add html-webpack-plugin -D
```

**webpack 配置**

```js
const HtmlWebpackPlugin = require('html-webpack-plugin')

module.exports = {
  // ...
  plugins: [
    new HtmlWebpackPlugin(),
  ],
}
```

这样配置好之后，构建时 `html-webpack-plugin` 会为我们创建一个 HTML 文件，其中会引用构建出来的 JS 文件。实际项目中，我们需要自己来写 HTML 文件，可以通过 `html-webpack-plugin` 的配置，传递一个写好的 HTML 模板：

```js
module.exports = {
  // ...
  plugins: [
    new HtmlWebpackPlugin({
      filename: 'index.html', // 配置输出文件名和路径
      template: 'assets/index.html', // 配置文件模板
    }),
  ],
}
```

> 如果需要添加多个页面关联，那么实例化多个 `html-webpack-plugin`，并将它们都放到 `plugins` 字段数组中就可以了。

### 构建 CSS

在配置中引入 loader 来解析和处理 CSS 文件：

```js
module.exports = {
  module: {
    rules: [
      // ...
      {
        test: /\.css/,
        include: [
          path.resolve(__dirname, 'src'),
        ],
        use: [
          'style-loader',
          'css-loader',
        ],
      },
    ],
  }
}
```

> `style-loader` 和 `css-loader` 都是单独的模块，需要安装。

- **css-loader** 负责解析 CSS 代码，主要是为了处理 CSS 中的依赖，例如 `@import` 和 `url()` 等引用外部文件的声明；
- **style-loader** 会将 css-loader 解析的结果转变成 JS 代码，运行时动态插入 style 标签来让 CSS 代码生效。

经由上述两个 `loader` 的处理后，CSS 代码会转变为 JS，和 `index.js` 一起打包了。如果需要单独把 CSS 文件分离出来，我们需要使用 [extract-text-webpack-plugin](https://webpack.docschina.org/plugins/extract-text-webpack-plugin) 插件。

> `extract-text-webpack-plugin` 这个插件如果在使用时并未发布支持 webpack 4.x 的正式版本，可以在安装的时候指定使用它的 alpha 版本：`npm install extract-text-webpack-plugin@next -D` 或者 `yarn add extract-text-webpack-plugin@next -D。如果你用的是 webpack 3.x 版本，直接用 extract-text-webpack-plugin` 现有的版本即可。

```js
const ExtractTextPlugin = require('extract-text-webpack-plugin')

module.exports = {
  // ...
  module: {
    rules: [
      {
        test: /\.css$/,
        // 因为这个插件需要干涉模块转换的内容，所以需要使用它对应的 loader
        use: ExtractTextPlugin.extract({ 
          fallback: 'style-loader',
          use: 'css-loader',
        }), 
      },
    ],
  },
  plugins: [
    // 引入插件，配置文件名，这里同样可以使用 [hash]
    new ExtractTextPlugin('index.css'),
  ],
}
```

### CSS 预处理器

以使用 Less 为例，可以在官方文档中找到对应的 [loader](https://webpack.docschina.org/loaders/less-loader)。

添加一个配置来支持解析后缀为 .less 的文件：

```js
module.exports = {
  // ...
  module: {
    rules: [
      {
        test: /\.less$/,
        // 因为这个插件需要干涉模块转换的内容，所以需要使用它对应的 loader
        use: ExtractTextPlugin.extract({ 
          fallback: 'style-loader',
          use: [
            'css-loader', 
            'less-loader',
          ],
        }), 
      },
    ],
  },
  // ...
}
```

### 处理图片文件

[file-loader](https://webpack.js.org/loaders/file-loader/) 可以用于处理很多类型的文件，它的主要作用是直接输出文件，把构建后的文件路径返回。配置很简单，在 `rules` 中添加一个字段，增加图片类型文件的解析配置：

```js
module.exports = {
  // ...
  module: {
    rules: [
      {
        test: /\.(png|jpg|gif)$/,
        use: [
          {
            loader: 'file-loader',
            options: {},
          },
        ],
      },
    ],
  },
}
```

### 使用 Babel

使用 ES6、ES7 标准来编写 JS 代码:

```js
module.exports = {
  // ...
  module: {
    rules: [
      {
        test: /\.jsx?/, // 支持 js 和 jsx
        include: [
          path.resolve(__dirname, 'src'), // src 目录下的才需要经过 babel-loader 处理
        ],
        loader: 'babel-loader',
      },
    ],
  },
}
```

Babel 的相关配置可以在目录下使用 .babelrc 文件来处理，详细参考 Babel 官方文档 [.babelrc](http://babeljs.io/docs/en/babelrc/)。

### 启动静态服务

我们可以使用 [webpack-dev-server](https://github.com/webpack/webpack-dev-server) 在本地开启一个简单的静态服务来进行开发。

```json
"scripts": {
  "build": "webpack --mode production",
  "start": "webpack-dev-server --mode development"
}
```

默认是访问 `index.html`，如果是其他页面要注意访问的 `URL` 是否正确。

## webpack 解析代码模块路径

webpack 依赖 [enhanced-resolve](https://github.com/webpack/enhanced-resolve/) 来解析代码模块的路径，webpack 配置文件中和 `resolve` 相关的选项都会传递给 `enhanced-resolve` 使用。

> 参考文章：[深入 Node.js 的模块机制](http://www.infoq.com/cn/articles/nodejs-module-mechanism)

### 模块解析规则

- 解析相对路径
  - 查找相对当前模块的路径下是否有对应文件或文件夹
  - 是文件则直接加载
  - 是文件夹则继续查找文件夹下的 package.json 文件
  - 有 package.json 文件则按照文件中 `main` 字段的文件名来查找文件
  - 无 package.json 或者无 `main` 字段则查找 `index.js` 文件
- 解析模块名
  - 查找当前文件目录下，父级目录及以上目录下的 `node_modules` 文件夹，看是否有对应名称的模块
- 解析绝对路径（不建议使用）
  - 直接查找对应路径的文件

在 webpack 配置中，和模块路径解析相关的配置都在 `resolve` 字段下：

```js
module.exports = {
  resolve: {
    // ...
  }
}
```

### 常用的一些配置

- resolve.alias
- resolve.extensions
- resolve.modules
- resolve.mainFiles
- resolve.resolveLoader

**`resolve.alias`**  模块别名

```js
alias: {
  utils: path.resolve(__dirname, 'src/utils') // 这里使用 path.resolve 和 __dirname 来获取绝对路径
  utils$: path.resolve(__dirname, 'src/utils')  // 精确匹配，只会匹配 import 'utils'
}
```

```js
import 'utils/query.js' // 等同于 import '[项目绝对路径]/src/utils/query.js'
```

> 参考文献： [Resolve Alias](https://webpack.docschina.org/configuration/resolve/#resolve-alias)

**`resolve.extensions`**

```js
extensions: ['.wasm', '.mjs', '.js', '.json', '.jsx'],
// 这里的顺序代表匹配后缀的优先级，例如对于 index.js 和 index.jsx，会优先选择 index.js
```

webpack 会尝试帮你补全那些后缀名来进行查找，例如有了上述的配置，当你在 `src/utils/` 目录下有一个 `common.js` 文件时，就可以这样来引用：

```js
import * as common from './src/utils/common'
```

- `resolve.modules` 对于直接声明依赖名的模块（如 `react` ），webpack 会类似 Node.js 一样进行路径搜索，搜索 `node_modules` 目录，这个目录就是使用 `resolve.modules` 字段进行配置的，默认就是：

```js
resolve: {
  modules: ['node_modules'],
},
```

通常情况下，我们不会调整这个配置，但是如果可以确定项目内所有的第三方依赖模块都是在项目根目录下的 `node_modules` 中的话，那么可以在 `node_modules` 之前配置一个确定的绝对路径：

```js
resolve: {
  modules: [
    path.resolve(__dirname, 'node_modules'), // 指定当前目录下的 node_modules 优先查找
    'node_modules', // 如果有一些类库是放在一些奇怪的地方的，你可以添加自定义的路径或者目录
  ],
},
```

> 这样配置在某种程度上可以简化模块的查找，提升构建速度。

**`resolve.mainFields`** 有 package.json 文件则按照文件中 `main` 字段的文件名来查找文件

当引用的是一个模块或者一个目录时，会使用 package.json 文件的哪一个字段下指定的文件，默认的配置是这样的：

```js
resolve: {
  // 配置 target === "web" 或者 target === "webworker" 时 mainFields 默认值是：
  mainFields: ['browser', 'module', 'main'],

  // target 的值为其他时，mainFields 默认值为：
  mainFields: ["module", "main"],
},
```

因为通常情况下，模块的 package 都不会声明 `browser` 或 `module` 字段，所以便是使用 `main` 了。

**`resolve.mainFiles`** 当目录下没有 package.json 文件时，我们说会默认使用目录下的 index.js 这个文件，其实这个也是可以配置的，默认配置是：

```js
resolve: {
  mainFiles: ['index'], // 你可以添加其他默认使用的文件名
},
```

通常情况下我们也无须修改这个配置。

**`resolve.resolveLoader`** 用于配置解析 `loader` 时的 `resolve` 配置，原本 `resolve` 的配置项在这个字段下基本都有。我们看下默认的配置：

```js
resolve: {
  resolveLoader: {
    extensions: ['.js', '.json'],
    mainFields: ['loader', 'main'],
  },
},
```

这里提供的配置相对少用，我们一般遵从标准的使用方式，使用默认配置，然后把 `loader` 安装在项目根路径下的 `node_modules` 下就可以了。

webpack 提供的这些选项可以更加灵活地去控制项目中代码模块的解析，除了上述的选项外，其他的选项在日常项目中相对比较少用到，如若需要，可以参考官方文档 [Resolve](https://webpack.docschina.org/configuration/resolve/)。

下一节：[配置loader](https://github.com/lsxlsxxslxsl/Read-Books-Notes/blob/master/post/webpck4%E5%88%9D%E4%BD%93%E9%AA%8C/02.%E9%85%8D%E7%BD%AEloader.md)